/*--------------------------------------------------------------------------*/
/*               The Opus Computer-Based Conversation System                */
/*       (c) Copyright 1986, Wynn Wagner III, All Rights Reserved           */
/*                                                                          */
/*         The original Sealink protocol is copyrighted by SEA, inc.        */
/*                                                                          */
/*                 This module was written by W.Wagner III                  */
/*        with modifications done by Bob Hartman and Rick Huebner           */
/*                                                                          */
/*                                                                          */
/*                                                                          */
/*         GENERAL PURPOSE FILE RECEPTION (Xmodem/YModem/Sealink)           */
/*                                                                          */
/*                                                                          */
/*  This module is similar to a routine used by Opus-Cbcs (1.00).  It is    */
/*  provided for your information only.  You will find routines that need   */
/*  to be coded and identifiers to be resolved.                             */
/*                                                                          */
/*  There is absolutely no guarantee that anything here will work.  If you  */
/*  break this routine, you own both pieces.                                */
/*                                                                          */
/*  USAGE:  You may use this material in any program with no obligation     */
/*          as long as there is no charge for your program.  For more       */
/*          information about commercial use, contact the "OPUSinfo HERE"   */
/*          BBS (124/111).                                                  */
/*                                                                          */
/*                                                                          */
/*--------------------------------------------------------------------------*/
#include <sys/types.h>
#include <sys/utime.h>
#include <ctype.h>
#include <stdio.h>
#include "xfer.h"
#include "com.h"

static void     pascal
SendACK (void);
static void     pascal
SendNAK (void);
static void     pascal
getblock (void);


/*--------------------------------------------------------------------------*/
/* EXTERNALS                                                                */
/*--------------------------------------------------------------------------*/
extern char    *UNLINK_msg;
extern byte    *local_CEOL;
extern int      remote_net, remote_node;
extern int      overwrite;
extern int      un_attended;
extern int      fullscreen;
extern long     timezone;
extern int      no_overdrive;
extern int      who_is_he;
extern struct _newnode newnodedes;     /* structure in new list     */

char           *ultoa ();


/*--------------------------------------------------------------------------*/
/* LOCALS                                                                   */
/*--------------------------------------------------------------------------*/
static unsigned int block_number;
static unsigned int base_block;
static char    *stat_msg;
static int      sliding;
static int      block_size;
static int      do_chksum;
static int      errs;
static int      real_errs;
static int      fsize1;
static int      fsize2;
static char    *buffer;
char           *malloc ();
static FILE    *infile;
static char    *_fpath;
static struct zero_block *header;
static char     final_name[80];
static int      first_block;    /* 1 = first block/Matrix bundle */

static char    *receiving = "Receiving  ";
static int      may_be_seadog;
static int      netmail;
static int      recv_ackless;
static int      did_nak;
static long     filesecs;

#define two_s "%s%s"



/*--------------------------------------------------------------------------*/
/* SEND ACK                                                                 */
/*--------------------------------------------------------------------------*/
static void     pascal
SendACK ()
{
   if ((!recv_ackless) || (block_number == 0))
      {
      SENDBYTE (ACK);
      if (sliding)
         {
         SENDBYTE (block_number);
         SENDBYTE (~block_number);
         }
      }

   if (block_number <= fsize1)
      {
      /* No-nonsense status line.  Don't waste time with frills... */
      gotoxy (locate_x, locate_y);
      cputs (ultoa (((unsigned long) (block_number)), e_input, 10));
      }

   errs = 0;
}                               /* SendACK */



/*--------------------------------------------------------------------------*/
/* SEND NAK                                                                 */
/*--------------------------------------------------------------------------*/
static void     pascal
SendNAK ()
{
   int             i;

   errs++;
   real_errs++;
   if (errs > 6)
      {
      stat_msg = FUBAR_msg;
      return;
      }

   if (++did_nak > 8)
      {
      recv_ackless = 0;
      }

   CLEAR_INBOUND ();

   /*--------------------------------------------------------------------*/
   /* Let the connection cool its heels.                                 */
   /*--------------------------------------------------------------------*/
   if (!recv_ackless)
      {
      if ((base_block != block_number) || (errs > 1))
         do
            {
            i = TIMED_READ (1);
            if (!CARRIER)
               return;
            }
         while (i >= 0);
      }

   if (block_number > base_block)
      SENDBYTE (NAK);
   else
      {
      if ((errs < 5) && (!do_chksum))
         SENDBYTE ('C');
      else
         {
         do_chksum = 1;
         SENDBYTE (NAK);
         }
      }

   if (sliding)
      {
      SENDBYTE (block_number);
      SENDBYTE (~block_number);
      }

   if (block_number <= fsize1)
      {
      /* No time for anything but the bare essentials... */
      gotoxy (locate_x, locate_y);
      cputs (ultoa (((unsigned long) (block_number)), e_input, 10));
      }
}                               /* SendNAK */

/*--------------------------------------------------------------------------*/
/* GET BLOCK                                                                */
/*--------------------------------------------------------------------------*/
static void     pascal
getblock ()
{
   register int    i;
   register char  *sptr;

   unsigned int    crc;
   int             loop;
   int             in_char;
   int             is_resend;
   int             blockerr;
   unsigned char   chksum;
   char            junkbuff[128];
   int             cur_x, cur_y;

   blockerr = loop = is_resend = 0;
   chksum = '\0';


   /*--------------------------------------------------------------------*/
   /* HEADER: Block number                                               */
   /*--------------------------------------------------------------------*/
   in_char = TIMED_READ (5);
   if (in_char != (block_number & 0xff))
      {
      if (in_char < block_number)
         is_resend = 1;
      else
      if ((block_number) || (in_char != 1))
         {
         blockerr++;
         stat_msg = SYNC_msg;
         }
      else
         block_number = 1;
      }

   /*--------------------------------------------------------------------*/
   /* HEADER: Complement of the block number                             */
   /*--------------------------------------------------------------------*/
   i = TIMED_READ (5);
   if ((i & 0xff) != ((~in_char) & 0xff))
      {
      blockerr++;
      stat_msg = CMPL_msg;
      }

   /*--------------------------------------------------------------------*/
   /* DATA                                                               */
   /*--------------------------------------------------------------------*/
   for (sptr = buffer, i = 0; i < block_size; i++, sptr++)
      {
      if (!CARRIER)
         return;

      in_char = TIMED_READ (5);
      if (in_char < 0)
         {
         SendNAK ();
         stat_msg = TIME_msg;
         return;
         }
      sptr[0] = (char) in_char;
      }

   /*--------------------------------------------------------------------*/
   /* CHECK                                                              */
   /*--------------------------------------------------------------------*/
   if (do_chksum)
      {
      for (sptr = buffer, i = 0; i < block_size; i++, sptr++)
         chksum += sptr[0];

      if (TIMED_READ (5) != chksum)
         {
         stat_msg = CHK_msg;
         blockerr++;
         }
      }
   else
      {
      unsigned int    lsb, msb;

      for (sptr = buffer, i = crc = 0; i < block_size; i++, sptr++)
         crc = xcrc (crc, (byte) sptr[0]);

      msb = TIMED_READ (3);
      lsb = TIMED_READ (3);
      if ((lsb < 0) || (msb < 0))
         {
         stat_msg = SHRT_msg;
         if (!block_number)
            sliding = 0;
         blockerr++;
         }
      else
      if (((msb << 8) | lsb) != crc)
         {
         stat_msg = CRC_msg;
         blockerr++;
         }
      }

   if (blockerr)
      SendNAK ();
   else
      {
      SendACK ();
      if (is_resend)
         return;

      if (block_number)
         {
         if (block_number < fsize1)
            fwrite (buffer, block_size, 1, infile);
         else
         if (block_number == fsize1)
            fwrite (buffer, fsize2, 1, infile);

         if (first_block)
            {
            struct _pkthdr *bundle;

            bundle = (struct _pkthdr *) buffer;
            remote_net = bundle->orig_net;
            remote_node = bundle->orig_node;
            first_block = 0;
            if (who_is_he)
               {
/* This should have the proper zone in it */
if (nodefind (ctl.our_zone, remote_net, remote_node))
                  {
                  sprintf (junkbuff, "Remote System: %s (%u/%u)",
                          newnodedes.SystemName,
                          remote_net,
                          remote_node);
                  }
               else
                  {
                  sprintf (junkbuff, "Remote System: UNKNOWN - LoTek Mailer (%u/%u)",
                          remote_net,
                          remote_node);
                  }

               if (un_attended && fullscreen)
                  {
                  cur_x = locate_x;
                  cur_y = locate_y;
                  gotoxy (0, 18);
                  set_xy (junkbuff);
                  gotoxy (cur_x, cur_y);
                  locate_x = cur_x;
                  locate_y = cur_y;
                  }
               who_is_he = 0;
               }
            }
         }
      else
         {
         if (!first_block)
            {
            sptr = &(header->name[0]);
            if (netmail)
               invent_pkt_name (sptr);

            for (i = 0; ((sptr[i]) && (i < 17)); i++)
               if (sptr[i] <= ' ')
                  sptr[i] = '\0';

            if (sptr[0])
               {
               sprintf (final_name, two_s, _fpath, sptr);
               sptr = final_name;
               fancy_str (sptr);

               i = strlen (final_name) - 1;

               if (un_attended && fullscreen)
                  {
                  clear_filetransfer ();
                  gotoxy (0, 13);
                  }
               else
               if (locate_y > 1)
                  {
                  gotoxy (0, locate_y - 1);
                  }
               cputs (sptr);
               }
            else
               status_line ("!Grunged hdr");
            }
         else
            message (NULL);

         if (header->size)
            {
            fsize1 = (int) (header->size / 128L);
            if (fsize2 = (header->size % 128))
               ++fsize1;
            else
               fsize2 = 128;
            cprintf ("%s %d blks", (netmail) ? "     " : "", fsize1);
            }

         if (header->time)
            {
            filesecs = header->time;
            }
         else
            {
            filesecs = -1;
            }

         if (header->moi[0])
            cprintf (" from %s", header->moi);

         if (header->noacks && !no_overdrive)
            {
            recv_ackless = 1;
            }

         if (!netmail)
            {
            if (un_attended && fullscreen)
               gotoxy (0, 13);
            set_xy (receiving);
            }
         }
      block_number++;
      }
}                               /* getblock */

/*--------------------------------------------------------------------------*/
/* GENERAL PURPOSE FILE RECEIPTION ROUTINE                                  */
/*--------------------------------------------------------------------------*/
char           *
receive_file (fpath, fname, protocol)
char           *fpath, *fname, protocol;
{
   char            tmpname[80];
   int             in_char;
   int             block_timer;
   struct utimbuf  times;
   long            t1, timerset ();

   did_nak = 0;
   filesecs = -1;
   brk_disable ();
   XON_DISABLE ();

   if (protocol == 'F')
      {
      may_be_seadog = 1;
      netmail = 0;
      protocol = 'S';
      first_block = 0;
      }
   else
   if (protocol == 'B')
      {
      protocol = 'S';
      may_be_seadog = 1;
      netmail = 1;
      first_block = 1;
      }
   else
      {
      may_be_seadog = 0;
      netmail = 0;
      first_block = 0;
      }

   CLEAR_IOERR ();
   fsize1 = 32767;
   if (fname)
      for (in_char = 0; fname[in_char]; in_char++)
         if ((fname[in_char] == '*') || (fname[in_char] == '?'))
            fname[0] = '\0';

   _fpath = fpath;
   sliding = 1;
   base_block =
      do_chksum =
      errs = 0;
   block_size = 128;

   locate_y = wherey ();
   locate_x = 3 + wherex ();

   switch (protocol)
      {
      case 'X':
         base_block = 1;
         sliding = 0;
         break;
      case 'Y':
         base_block = 1;
         sliding = 0;
         block_size = 1024;
         break;
      case 'S':
         break;
      case 'T':
         break;
      case 'M':
         base_block = 1;
         sliding = 0;
         break;
      default:
         cprintf ("Protocol??\r\n");
         return NULL;
      }

   block_number = base_block;

   sprintf (tmpname, "%s_TMP_.$$$", fpath);
   sprintf (final_name, two_s, _fpath, (fname && fname[0]) ? fname : "UNKNOWN.$$$");

   infile = fopen (tmpname, write_binary);
   if (got_error (CREATE_msg, tmpname))
      return (NULL);
   if (isatty (fileno (infile)))
      {
      errno = 1;
      got_error (DEVICE_msg, tmpname);
      fclose (infile);
      return (NULL);
      }

   buffer = malloc (1032);
   header = (struct zero_block *) buffer;
   if (!buffer)
      {
      status_line ("!MEM:ulbuf");
      return (NULL);
      }

   if (un_attended && fullscreen)
      {
      gotoxy (0, 13);
      }
   set_xy (receiving);
   stat_msg = NULL;
   throughput (0, 0L);

   /* If we did the SEAdog kludge, then don't send another NAK */
   if (!may_be_seadog)
      SendNAK ();

   t1 = timerset (300);
   real_errs = 0;
   do
      {
loop_top:
      if (((KEYPRESS ()) && (READKB () == 27)))
         {
         stat_msg = KBD_msg;
         goto fubar;
         }

      in_char = TIMED_READ (4);

      switch (in_char)
         {
         case SOH:
            /*--------------------------------------------*/
            /* SOH: 128 byte block header                 */
            /*--------------------------------------------*/
            block_size = 128;
            getblock ();
            t1 = timerset (300);
            break;

         case STX:
            /*--------------------------------------------*/
            /* STX: YModem 1k block header                */
            /*--------------------------------------------*/
            block_size = 1024;
            getblock ();
            t1 = timerset (300);
            break;

         case SYN:
            /*--------------------------------------------*/
            /* SYN: TeLink Zero block header              */
            /*--------------------------------------------*/
            do_chksum = 1;
            getblock ();
            do_chksum = 0;
            t1 = timerset (300);
            break;

         case CAN:
            /*--------------------------------------------*/
            /* CAN                                        */
            /*--------------------------------------------*/
            if (TIMED_READ (2) == CAN)
               {
               stat_msg = CAN_msg;
               goto fubar;
               }
            t1 = timerset (300);
            break;

         case EOT:
            /*--------------------------------------------*/
            /* EOT                                        */
            /*--------------------------------------------*/
            stat_msg = EOT_msg;
            t1 = timerset (20);
            while (!timeup (t1))
               {
               TIMED_READ (0);
               time_release ();
               }
            if (block_number)
               goto done;
            else
               goto fubar;

         default:
            /*--------------------------------------------*/
            /* Default                                    */
            /*--------------------------------------------*/
            if (in_char > 0)
               {
               if (!may_be_seadog)
                  {
                  gotoxy (locate_x, locate_y);
                  cprintf ("??? [%Xh, blk %d] ", in_char, block_number);
                  set_xy (receiving);
                  }
               }

            /*
             * Debris is tolerable for up to the amount of time we are
             * willing to wait for a start of a block.  Right now it is 10
             * seconds. 
             */

            /*
             * If may_be_seadog is set, then we know we should get an SOH
             * pretty soon 
             */
            if (timeup (t1) || !may_be_seadog)
               SendNAK ();
            break;
         }                      /* switch */

      if (errs > 14)
         {
         stat_msg = FUBAR_msg;
         goto fubar;
         }

      else
      if (got_error (WRITE_msg, fname))
         {
         stat_msg = IO_msg;
         goto fubar;
         }

      if (stat_msg)
         {
         cputs (stat_msg);
         cputs (local_CEOL);
         stat_msg = NULL;
         }
      } while (CARRIER);

   stat_msg = CARRIER_msg;

fubar:

   if (buffer)
      free (buffer);
   if (infile)
      {
      fclose (infile);
      got_error (CLOSE_msg, tmpname);
      unlink (tmpname);
      got_error (UNLINK_msg, tmpname);
      }

   CLEAR_OUTBOUND ();

   if (stat_msg != EOT_msg)
      {
      for (in_char = 0; in_char < 5; in_char++)
         SENDBYTE (CAN);

      status_line ("!File not received: %s", stat_msg);
      }

   CLEAR_INBOUND ();
   return NULL;


done:

   recv_ackless = 0;
   SendACK ();

   if (buffer)
      free (buffer);

   if (infile)
      {
      int             i, j, k;
      struct FILEINFO dta;

      fclose (infile);
      got_error (CLOSE_msg, tmpname);

      i = strlen (tmpname) - 1;
      j = strlen (final_name) - 1;

      if (tmpname[i] == '.')
         tmpname[i] = '\0';
      if (final_name[j] == '.')
         {
         final_name[j] = '\0';
         --j;
         }

      i = 0;
      k = is_arcmail (final_name, j);
      if ((!overwrite) || k)
         {
         while (rename (tmpname, final_name))
            {
            if (isdigit (final_name[j]))
               final_name[j]++;
            else
               final_name[j] = '0';
            if (!isdigit (final_name[j]))
               return (tmpname);
            i = 1;
            }
         CLEAR_IOERR ();
         }
      else
         {
         unlink (final_name);
         rename (tmpname, final_name);
         }
      if (i)
         {
         if (locate_y)
            gotoxy (0, locate_y - 1);
         status_line ("+Dupe file renamed: %s", final_name);
         }

      if (filesecs != -1)
         {
         tzset ();
         times.modtime = filesecs + timezone;
         times.actime = filesecs + timezone;
         utime (final_name, &times);
         }

      if (!dfind (&dta, final_name, 0))
         {
         if (!i)
            {
            if (locate_y)
               gotoxy (0, locate_y - 1);
            }
         if (real_errs > 4)
            status_line ("+Corrected %d errors in %d blocks",
                         real_errs, fsize1);
         status_line ("+Received-S %s", final_name);
         throughput (1, dta.size);

         dta.size /= 1024L;
         strcpy (final_name, dta.name);
         CLEAR_IOERR ();
         return final_name;     /* signal what file we got */
         }
      }
   CLEAR_IOERR ();
   return NULL;

}

/* END OF FILE: recv_sl.c */
