/*--------------------------------------------------------------------------*/
/*                                                                          */
/*                                                                          */
/*      ------------         Bit-Bucket Software <no-Inc>                   */
/*      \ 10001101 /         Writers and Distributors of                    */
/*       \ 011110 /          No-Cost<no-tm> Software.                       */
/*        \ 1011 /                                                          */
/*         ------            KopyRong (K) 1987.  ALL RIGHTS REVERSED.       */
/*                                                                          */
/*                                                                          */
/*               This module was written by Vince Perriello                 */
/*                                                                          */
/*                                                                          */
/*                 BinkleyTerm Nodelist processing module                   */
/*                                                                          */
/*                                                                          */
/*   This  software  package is being distributed WITH FULL SOURCE CODE     */
/*   with the  following  conditions:    1)  If  anything awful happens     */
/*   because  you  use    it   (or  don't  use  it),  you  accept  full     */
/*   responsibility;  2) you  don't start making tons of voice calls to     */
/*   the authors to complain or  make  suggestions  about enhancements,     */
/*   useful or otherwise;  3) you  do not reuse this code in commercial     */
/*   products without specific permission to do so  from  the  authors;     */
/*   4) If you find any problems you send  fixes  to  the  authors  for     */
/*   inclusion  in  updates;    5) You find some way  to  express  your     */
/*   appreciation  for  this  method of distribution, either by writing     */
/*   code or  application  notes,  or  just sending along a "Thank You"     */
/*   message.                                                               */
/*                                                                          */
/*   There is  copyrighted  code  in  this product.  We either wrote it     */
/*   ourselves or got  permission  to use it.  Please don't force us to     */
/*   pay a lawyer --  have some respect for our motives and don't abuse     */
/*   this "license".                                                        */
/*                                                                          */
/*                                                                          */
/*--------------------------------------------------------------------------*/

#include <sys\types.h>
#include <sys\stat.h>
#include <stdio.h>
#include <fcntl.h>
#include <signal.h>
#include <ctype.h>
#include <conio.h>
#include "com.h"
#include "xfer.h"
#include "keybd.h"

int             newnodelist = 0;/* if 1, use new nodelist.   */
extern struct pointers ctl;     /* where all the strings is  */
int             autobaud = 0;   /* Use highest baud rate when calling out */
struct _newnode newnodedes;     /* structure in new list     */
static struct _ndi *node_index = NULL;  /* pointer to node array     */
static int      idx_size = 0;   /* number of entries in it   */
extern int      boss_net, boss_node;    /* BOSS net and node         */
extern char    *BOSSphone, *BOSSpwd;    /* BOSS phone no. & password */

/*---------------------------------------------------------------------------*/
/* NODEPROC                                                                  */
/* Find nodelist entry and set baud to nodelist baud for dialing out         */
/*---------------------------------------------------------------------------*/

nodeproc (nodeaddr)
char           *nodeaddr;
{
   int             opuszone, opusnet, opusnode;   /* zone, net and node */
   int             i;
   char           *c, *skip_blanks ();

   c = skip_blanks (nodeaddr);  /* get rid of the blanks     */
   if (sscanf (c, "%d:%d/%d", &opuszone, &opusnet, &opusnode) != 3)
      {
      opuszone = ctl.our_zone;
      if (sscanf (c, "%d/%d", &opusnet, &opusnode) != 2)
         {
         status_line ("!Can't convert Net/Node address");
         return (0);
         }
      }
   if (!nodefind (opuszone, opusnet, opusnode))   /* if we can't find the node */
      return (0);               /* go away now               */
   if (opuszone != ctl.our_zone)
      {
      status_line ("*Processing node %d:%d/%d -- %s", opuszone, opusnet, opusnode, newnodedes.SystemName);
      }
   else
      {
      status_line ("*Processing node %d/%d -- %s", opusnet, opusnode, newnodedes.SystemName);
      }
   if (!CARRIER)                /* if no carrier yet,        */
      {
      if (autobaud)
         set_baud (ctl.max_baud, 1);    /* Set to our highest baud rate */
      else
         set_baud ((i = 300 * newnodedes.BaudRate), 1); /* set baud to nodelist
                                                         * baud */
      }
   return (1);                  /* return success to caller  */
}

/*---------------------------------------------------------------------------*/
/* NODEFIND                                                                  */
/* Find nodelist entry for use by other routines (password, nodeproc)        */
/* If found, result will be in "newnodedes".                                 */
/*---------------------------------------------------------------------------*/

nodefind (opuszone, opusnet, opusnode)
int             opuszone, opusnet, opusnode;      /* zone, net and node */
{
   register struct _ndi *nodeidx;       /* index file                */
   int             foundnet = 0;/* 'found the net' flag      */
   int             found = 0;   /* 'found the node' flag     */
   int             nodeoff = 0; /* offset into nodelist.sys  */
   char            temp[80];    /* where we build filenames  */
   int             i;
   FILE           *stream;
   struct stat     f;
   int             have_boss_data = 0;
   int             proper_zone = 0;

   newnodedes.NetNumber = newnodedes.NodeNumber = 0;

   if ((opusnet == boss_net) && (opusnode == boss_node) &&
       (opuszone == ctl.our_zone))
      {
      if (strlen (BOSSphone) > 2)
         ++have_boss_data;
      if (strlen (BOSSpwd) > 2)
         ++have_boss_data;
      }

   if (node_index == NULL)
      {
      temp[0] = '\0';           /* "null-terminated string"  */
      strcpy (temp, ctl.net_info);      /* take nodelist path        */
      strcat (temp, "NODELIST.IDX");    /* add in the file name      */
      if ((stream = fopen (temp, "rb")) == NULL)        /* OK, let's open the
                                                         * file   */
         {
         i = 0;                 /* Need this later           */
         if (have_boss_data == 2)
            goto fill_in_boss;
         status_line ("!Unable to open Nodelist.Idx file");
         return (0);            /* no file, no work to do    */
         }
      fstat (fileno (stream), &f);      /* get file statistics       */
      i = f.st_size;            /* size of index file,       */
      idx_size = i / sizeof (struct _ndi);      /* number of index entries   */
      node_index = (struct _ndi *) malloc (i);
      if (node_index == NULL)
         {
         status_line ("!Unable to allocate memory for Nodelist Index");
         fclose (stream);
         return (0);
         }
      if (fread (node_index, i, 1, stream) != 1)
         {
         status_line ("!Failed to read nodelist index into memory");
         return (0);
         }
      fclose (stream);
      }
   nodeidx = node_index;

   /* If using the old nodelist, bypass zone stuff - it isn't there */
   if (!newnodelist)
      opuszone = 0;

/* This should go away when we really have it right */
opuszone = 0;
   for (i = 1; i <= idx_size; nodeoff++, nodeidx++, i++)
      {
      if (opuszone > 0)
         {
         if (nodeidx->node == -2)
            {
            if (proper_zone)
               break;
            if (nodeidx->net == opuszone)
               proper_zone = 1;
            }

         if (!proper_zone)
            continue;
         }

      if (nodeidx->net == opusnet)      /* if a match on net,        */
         {
         foundnet = 1;          /* say we found the net      */
         if (((opusnode == 0) && (nodeidx->node <= 0))
             || (nodeidx->node == opusnode))    /* see if we found the node  */
            {
            found = 1;          /* say we found it           */
            break;              /* get out                   */
            }
         }
      else
      if (foundnet)             /* already past the net?     */
         break;                 /* Yes, we failed...         */
      }

   if (!found)
      {
      if (have_boss_data == 2)
         goto fill_in_boss;
      status_line ("!Couldn't find Net/Node: %d/%d", opusnet, opusnode);
      return (0);
      }

   if (newnodelist)
      i = (get_new_info (nodeoff));
   else
      i = (get_old_info (nodeoff));

   if (!have_boss_data)
      return (i);

/*
 *	We can get here one of two ways:
 *
 *	1) No nodelist file was found, we branched directly
 *
 *	2) Nodelist lookup occurred, but this is the BOSS.
 *
 *	For case (1), have_boss_data MUST be 2 (meaning we have
 *	both a phone number and a password entry). If that is the
 *	case, fill in newnodedes with mostly zeroes, plugging in
 *	the BOSS net, node, phone number and password.
 *
 *	For case (2), just see if there is any substitution for
 *	BOSSphone and/or BOSSpwd, then exit.
 *
 */

fill_in_boss:

   if (strlen (BOSSphone) > 2)
      {
      strncpy (newnodedes.PhoneNumber, BOSSphone, 40);
      newnodedes.PhoneNumber[39] = '\0';
      }
   if (strlen (BOSSpwd) > 2)
      {
      strncpy (newnodedes.Password, BOSSpwd, 8);
      newnodedes.Password[7] = '\0';
      }
   if (i)
      return (1);

   /* No BOSS in the nodelist */
   if (have_boss_data != 2)
      return (0);

   newnodedes.NodeNumber = opusnode;    /* Node Number */
   newnodedes.NetNumber = opusnet;      /* Net Number  */
   newnodedes.Cost = newnodedes.RealCost = 0;   /* Assume boss is free */
   strcpy (newnodedes.SystemName, "Binkley's Boss");    /* System Name defaults */
   strcpy (newnodedes.MiscInfo, "Somewhere out There"); /* As does City */
   newnodedes.HubNode = 0;      /* Don't know who is HUB */
   newnodedes.BaudRate = ctl.max_baud / 300;    /* Assume boss speed = ours
                                                 * newnodedes.ModemType = 0;
                                                 * * Or modem type */
   newnodedes.NodeFlags = B_CM; /* Assume boss is CM */
   newnodedes.NodeFiller = 0;   /* Zero out filler */
   return (1);
}

int 
get_old_info (recno)
unsigned        recno;
{
   struct _node    nodedes;     /* desc. of node             */
   long            nodeoff;     /* Offset into NODELIST.SYS  */
   char            temp[80];    /* where we build filenames  */
   char           *c, *ch;
   int             i;
   FILE           *stream;

   nodeoff = (long) recno *(long) sizeof (nodedes);     /* actual file offset        */

   strcpy (temp, ctl.net_info); /* take nodelist path        */
   strcat (temp, "NODELIST.SYS");       /* add in the file name      */
   if ((stream = fopen (temp, "rb")) == NULL)   /* OK, let's open the file   */
      {
      status_line ("!Unable to open Nodelist.Sys file");
      return (0);
      }

   if (fseek (stream, nodeoff, SEEK_SET))       /* try to point at record    */
      {
      status_line ("!Unable to position to node record in Nodelist.Sys");
      fclose (stream);
      return (0);
      }

   if (!fread (&nodedes, sizeof (nodedes), 1, stream))
      {
      status_line ("!Not able to read node record from Nodelist.Sys");
      fclose (stream);
      return (0);
      }
   fclose (stream);
   /*
    * Copy data from nodedes into newnodedes.
   */

   newnodedes.NodeNumber = nodedes.number;      /* Node Number  */
   newnodedes.NetNumber = nodedes.net;  /* Net Number   */
   newnodedes.Cost = nodedes.cost;      /* Cost         */

   strncpy (&newnodedes.SystemName[0], &nodedes.name[0], 20);
   newnodedes.SystemName[19] = '\0';    /* System Name  */

   strncpy (&newnodedes.PhoneNumber[0], &nodedes.phone[0], 40);
   newnodedes.PhoneNumber[39] = '\0';   /* Phone Number */

   strncpy (&newnodedes.MiscInfo[0], &nodedes.city[0], 30);
   newnodedes.MiscInfo[29] = '\0';

   ch = NULL;                   /* Pointer to password      */
   c = nodedes.city;            /* Point to start of city   */
   i = 37;                      /* No password if this = 0  */
   while (i--)                  /* Enforce that limit       */
      {
      if (*c++ != '\0')         /* If not at end of city,   */
         continue;              /* go on to next character  */
      if (*c++ != '!')          /* End of city, got '!' ??  */
         break;                 /* No, treat like a failure */
      ch = c;                   /* Got it, point to it      */
      break;                    /* Exit with success code   */
      }
   if (ch != NULL)
      {
      strncpy (&newnodedes.Password[0], ch, 8); /* Copy the password        */
      }
   else
      newnodedes.Password[0] = '\0';    /* Else set zero length     */

   newnodedes.RealCost = nodedes.cost;  /* Cost                     */
   newnodedes.HubNode = 0;      /* Don't know who is Hub    */
   newnodedes.BaudRate = nodedes.rate / 300;    /* Baud Rate                */
   newnodedes.ModemType = 0;    /* Don't know modem type    */
   newnodedes.NodeFlags = 0;    /* Don't know any flags     */
   newnodedes.NodeFiller = 0;   /* Filler should be zero    */
   return (1);
}

int 
get_new_info (recno)
unsigned        recno;
{
   long            nodeoff;     /* Offset into NODELIST.DAT  */
   char            temp[80];    /* where we build filenames  */
   int             i;
   FILE           *stream;

   nodeoff = (long) recno *(long) sizeof (newnodedes);  /* actual file offset        */

   strcpy (temp, ctl.net_info); /* take nodelist path        */
   strcat (temp, "NODELIST.DAT");       /* add in the file name      */
   if ((stream = fopen (temp, "rb")) == NULL)   /* OK, let's open the file   */
      {
      status_line ("!Unable to open Nodelist.Dat file");
      return (0);
      }

   if (fseek (stream, nodeoff, SEEK_SET))       /* try to point at record    */
      {
      status_line ("!Unable to position to node record in Nodelist.Dat");
      fclose (stream);
      return (0);
      }

   if (!fread (&newnodedes, sizeof (newnodedes), 1, stream))
      {
      status_line ("!Not able to read node record from Nodelist.Dat");
      fclose (stream);
      return (0);
      }
   fclose (stream);
   return (1);
}
